@@ -65,6 +65,10 @@ struct join_message {
 	} nodes[SD_MAX_NODES];
 };
 
+struct leave_message {
+	struct message_header header;
+};
+
 struct vdi_op_message {
 	struct message_header header;
 	struct sd_vdi_req req;
@@ -98,7 +102,7 @@ struct work_confchg {
 	char name[128];						\
 	list_for_each_entry(node, node_list, list) {		\
 		dprintf("%c nodeid: %x, pid: %d, ip: %s\n",	\
-			is_myself(&node->ent) ? 'l' : ' ',	\
+			is_myself(node->ent.addr, node->ent.port) ? 'l' : ' ',	\
 			node->nodeid, node->pid,		\
 			addr_to_str(name, sizeof(name),		\
 			node->ent.addr, node->ent.port));	\
@@ -130,18 +134,6 @@ CPG_EVENT_WORK_FNS(RUNNING, running)
 CPG_EVENT_WORK_FNS(SUSPENDED, suspended)
 CPG_EVENT_WORK_FNS(JOINING, joining)
 
-static int node_cmp(const void *a, const void *b)
-{
-	const struct sheepdog_node_list_entry *node1 = a;
-	const struct sheepdog_node_list_entry *node2 = b;
-
-	if (node1->id < node2->id)
-		return -1;
-	if (node1->id > node2->id)
-		return 1;
-	return 0;
-}
-
 static int send_message(cpg_handle_t handle, struct message_header *msg)
 {
 	struct iovec iov;
@@ -198,9 +190,26 @@ int get_ordered_sd_node_list(struct sheepdog_node_list_entry *entries)
 	return build_node_list(&sys->sd_node_list, entries);
 }
 
-int setup_ordered_sd_node_list(struct request *req)
+void get_ordered_sd_vnode_list(struct sheepdog_vnode_list_entry *entries,
+			       int *nr_vnodes, int *nr_nodes)
 {
-	return get_ordered_sd_node_list(req->entry);
+	struct sheepdog_node_list_entry nodes[SD_MAX_NODES];
+	int nr;
+
+	nr = build_node_list(&sys->sd_node_list, nodes);
+	*nr_nodes = nr;
+
+	if (sys->nr_vnodes == 0)
+		sys->nr_vnodes = nodes_to_vnodes(nodes, nr, sys->vnodes);
+
+	memcpy(entries, sys->vnodes, sizeof(*entries) * sys->nr_vnodes);
+
+	*nr_vnodes = sys->nr_vnodes;
+}
+
+void setup_ordered_sd_vnode_list(struct request *req)
+{
+	get_ordered_sd_vnode_list(req->entry, &req->nr_vnodes, &req->nr_nodes);
 }
 
 static void get_node_list(struct sd_node_req *req,
@@ -330,7 +339,7 @@ static int is_master(void)
 		return 0;
 
 	node = list_first_entry(&sys->sd_node_list, struct node, list);
-	if (is_myself(&node->ent))
+	if (is_myself(node->ent.addr, node->ent.port))
 		return 1;
 	return 0;
 }
@@ -345,6 +354,7 @@ static int get_cluster_status(struct sheepdog_node_list_entry *from,
 	struct sheepdog_node_list_entry local_entries[SD_MAX_NODES];
 	struct node *node;
 	uint32_t local_epoch;
+	char str[256];
 
 	*status = sys->status;
 	if (inc_epoch)
@@ -359,37 +369,43 @@ static int get_cluster_status(struct sheepdog_node_list_entry *from,
 			break;
 
 		if (ctime != get_cluster_ctime()) {
-			eprintf("joining node has invalid ctime, %"PRIu64"\n", from->id);
+			eprintf("joining node has invalid ctime, %s\n",
+				addr_to_str(str, sizeof(str), from->addr, from->port));
 			return SD_RES_INVALID_CTIME;
 		}
 
 		local_epoch = get_latest_epoch();
 		if (epoch > local_epoch) {
-			eprintf("sheepdog is running with older epoch, %"PRIu32" %"PRIu32" %"PRIu64"\n",
-				epoch, local_epoch, from->id);
+			eprintf("sheepdog is running with older epoch, %"PRIu32" %"PRIu32" %s\n",
+				epoch, local_epoch,
+				addr_to_str(str, sizeof(str), from->addr, from->port));
 			return SD_RES_OLD_NODE_VER;
 		}
 		break;
 	case SD_STATUS_WAIT_FOR_FORMAT:
 		if (nr_entries != 0) {
-			eprintf("joining node is not clean, %"PRIu64"\n", from->id);
+			eprintf("joining node is not clean, %s\n",
+				addr_to_str(str, sizeof(str), from->addr, from->port));
 			return SD_RES_NOT_FORMATTED;
 		}
 		break;
 	case SD_STATUS_WAIT_FOR_JOIN:
 		if (ctime != get_cluster_ctime()) {
-			eprintf("joining node has invalid ctime, %"PRIu64"\n", from->id);
+			eprintf("joining node has invalid ctime, %s\n",
+				addr_to_str(str, sizeof(str), from->addr, from->port));
 			return SD_RES_INVALID_CTIME;
 		}
 
 		local_epoch = get_latest_epoch();
 		if (epoch > local_epoch) {
-			eprintf("sheepdog is waiting with older epoch, %"PRIu32" %"PRIu32" %"PRIu64"\n",
-				epoch, local_epoch, from->id);
+			eprintf("sheepdog is waiting with older epoch, %"PRIu32" %"PRIu32" %s\n",
+				epoch, local_epoch,
+				addr_to_str(str, sizeof(str), from->addr, from->port));
 			return SD_RES_OLD_NODE_VER;
 		} else if (epoch < local_epoch) {
-			eprintf("sheepdog is waiting with newer epoch, %"PRIu32" %"PRIu32" %"PRIu64"\n",
-				epoch, local_epoch, from->id);
+			eprintf("sheepdog is waiting with newer epoch, %"PRIu32" %"PRIu32" %s\n",
+				epoch, local_epoch,
+				addr_to_str(str, sizeof(str), from->addr, from->port));
 			return SD_RES_NEW_NODE_VER;
 		}
 
@@ -398,13 +414,15 @@ static int get_cluster_status(struct sheepdog_node_list_entry *from,
 		nr_local_entries /= sizeof(local_entries[0]);
 
 		if (nr_entries != nr_local_entries) {
-			eprintf("joining node has invalid epoch, %"PRIu32" %"PRIu64"\n",
-				epoch, from->id);
+			eprintf("joining node has invalid epoch, %"PRIu32" %s\n",
+				epoch,
+				addr_to_str(str, sizeof(str), from->addr, from->port));
 			return SD_RES_INVALID_EPOCH;
 		}
 
 		if (memcmp(entries, local_entries, sizeof(entries[0]) * nr_entries) != 0) {
-			eprintf("joining node has invalid epoch, %"PRIu64"\n", from->id);
+			eprintf("joining node has invalid epoch, %s\n",
+				addr_to_str(str, sizeof(str), from->addr, from->port));
 			return SD_RES_INVALID_EPOCH;
 		}
 
@@ -417,10 +435,10 @@ static int get_cluster_status(struct sheepdog_node_list_entry *from,
 			return SD_RES_SUCCESS;
 
 		for (i = 0; i < nr_local_entries; i++) {
-			if (local_entries[i].id == from->id)
+			if (node_cmp(local_entries + i, from) == 0)
 				goto next;
 			list_for_each_entry(node, &sys->sd_node_list, list) {
-				if (local_entries[i].id == node->ent.id)
+				if (node_cmp(local_entries + i, &node->ent) == 0)
 					goto next;
 			}
 			return SD_RES_SUCCESS;
@@ -487,7 +505,7 @@ static void get_vdi_bitmap_from_all(void)
 	nr_nodes = get_ordered_sd_node_list(entry);
 
 	for (i = 0; i < nr_nodes; i++) {
-		if (is_myself(&entry[i]))
+		if (is_myself(entry[i].addr, entry[i].port))
 			continue;
 
 		addr_to_str(host, sizeof(host), entry[i].addr, 0);
@@ -530,11 +548,11 @@ static int move_node_to_sd_list(uint32_t nodeid, uint32_t pid,
 	if (!node)
 		return 1;
 
-	if (!node->ent.id)
-		node->ent = ent;
+	node->ent = ent;
 
 	list_del(&node->list);
 	list_add_tail(&node->list, &sys->sd_node_list);
+	sys->nr_vnodes = 0;
 
 	return 0;
 }
@@ -546,7 +564,7 @@ static void update_cluster_info(struct join_message *msg)
 	struct sheepdog_node_list_entry entry[SD_MAX_NODES];
 
 	if (msg->result != SD_RES_SUCCESS) {
-		if (is_myself(&msg->header.from)) {
+		if (is_myself(msg->header.from.addr, msg->header.from.port)) {
 			eprintf("failed to join sheepdog, %d\n", msg->result);
 			sys->status = SD_STATUS_JOIN_FAILED;
 		}
@@ -737,7 +755,6 @@ static void vdi_op_done(struct vdi_op_message *msg)
 			eprintf("can't write epoch %u\n", sys->epoch);
 		update_epoch_store(sys->epoch);
 
-		set_nodeid(sys->this_node.id);
 		set_global_nr_copies(sys->nr_sobjs);
 
 		sys->status = SD_STATUS_OK;
@@ -750,7 +767,7 @@ static void vdi_op_done(struct vdi_op_message *msg)
 		ret = SD_RES_UNKNOWN;
 	}
 out:
-	if (!is_myself(&msg->header.from))
+	if (!is_myself(msg->header.from.addr, msg->header.from.port))
 		return;
 
 	req = list_first_entry(&sys->pending_list, struct request, pending_list);
@@ -795,8 +812,7 @@ static void __sd_deliver(struct cpg_event *cevent)
 			return;
 		}
 
-		if (!node->ent.id)
-			node->ent = m->from;
+		node->ent = m->from;
 	}
 
 	if (m->state == DM_INIT && is_master()) {
@@ -811,17 +827,6 @@ static void __sd_deliver(struct cpg_event *cevent)
 			break;
 		}
 	}
-
-	if (m->state == DM_FIN) {
-		switch (m->op) {
-		case SD_MSG_JOIN:
-			update_cluster_info((struct join_message *)m);
-			break;
-		default:
-			eprintf("unknown message %d\n", m->op);
-			break;
-		}
-	}
 }
 
 static void send_join_response(struct work_deliver *w)
@@ -840,9 +845,44 @@ static void __sd_deliver_done(struct cpg_event *cevent)
 	struct message_header *m;
 	char name[128];
 	int do_recovery;
+	struct node *node;
+	struct sheepdog_node_list_entry e[SD_MAX_NODES];
+	int nr;
 
 	m = w->msg;
-	do_recovery = (m->state == DM_FIN && m->op == SD_MSG_JOIN);
+
+	if (m->state == DM_FIN) {
+		switch (m->op) {
+		case SD_MSG_JOIN:
+			update_cluster_info((struct join_message *)m);
+			break;
+		case SD_MSG_LEAVE:
+			node = find_node(&sys->sd_node_list, m->nodeid, m->pid);
+			if (node) {
+				sys->nr_vnodes = 0;
+
+				list_del(&node->list);
+				free(node);
+				if (sys->status == SD_STATUS_OK) {
+					nr = get_ordered_sd_node_list(e);
+					dprintf("update epoch, %d, %d\n", sys->epoch + 1, nr);
+					epoch_log_write(sys->epoch + 1, (char *)e,
+							nr * sizeof(struct sheepdog_node_list_entry));
+
+					sys->epoch++;
+
+					update_epoch_store(sys->epoch);
+				}
+			}
+			break;
+		default:
+			eprintf("unknown message %d\n", m->op);
+			break;
+		}
+	}
+
+	do_recovery = (m->state == DM_FIN &&
+		       (m->op == SD_MSG_JOIN || m->op == SD_MSG_LEAVE));
 
 	dprintf("op: %d, state: %u, size: %d, from: %s\n",
 		m->op, m->state, m->msg_length,
@@ -940,6 +980,7 @@ static void del_node(struct cpg_address *addr, struct work_confchg *w)
 		struct sheepdog_node_list_entry e[SD_MAX_NODES];
 
 		w->sd_node_left++;
+		sys->nr_vnodes = 0;
 
 		list_del(&node->list);
 		free(node);
@@ -1213,8 +1254,13 @@ out:
 	cpg_event_free(cevent);
 	cpg_event_clear_running();
 
-	if (!list_empty(&sys->cpg_event_siblings) && !cpg_event_suspended())
-		start_cpg_event_work();
+	if (!list_empty(&sys->cpg_event_siblings)) {
+		if (cpg_event_joining())
+			/* io requests need to return SD_RES_NEW_NODE_VER */
+			start_cpg_event_work();
+		else if (!cpg_event_suspended())
+			start_cpg_event_work();
+	}
 }
 
 static int check_epoch(struct request *req)
@@ -1254,6 +1300,20 @@ int is_access_to_busy_objects(uint64_t oid)
 	return 0;
 }
 
+static int __is_access_to_recoverying_objects(struct request *req)
+{
+	if (req->rq.flags & SD_FLAG_CMD_RECOVERY) {
+		if (req->rq.opcode != SD_OP_READ_OBJ)
+			eprintf("bug\n");
+		return 0;
+	}
+
+	if (is_recoverying_oid(req->local_oid))
+		return 1;
+
+	return 0;
+}
+
 static int __is_access_to_busy_objects(struct request *req)
 {
 	if (req->rq.flags & SD_FLAG_CMD_RECOVERY) {
@@ -1265,9 +1325,6 @@ static int __is_access_to_busy_objects(struct request *req)
 	if (is_access_to_busy_objects(req->local_oid))
 		return 1;
 
-	if (is_recoverying_oid(req->local_oid))
-		return 1;
-
 	return 0;
 }
 
@@ -1304,6 +1361,21 @@ void start_cpg_event_work(void)
 	if (cpg_event_joining()) {
 		if (!cpg_event_suspended())
 			panic("should not happen\n");
+
+		if (cevent->ctype == CPG_EVENT_REQUEST) {
+			struct request *req = container_of(cevent, struct request, cev);
+			if (is_io_request(req->rq.opcode) && req->rq.flags & SD_FLAG_CMD_DIRECT) {
+				list_del(&cevent->cpg_event_list);
+
+				req->rp.result = SD_RES_NEW_NODE_VER;
+
+				/* TODO: cleanup */
+				list_add_tail(&req->r_wlist, &sys->outstanding_req_list);
+				sys->nr_outstanding_io++;
+
+				req->work.done(&req->work, 0);
+			}
+		}
 		return;
 	}
 
@@ -1319,6 +1391,20 @@ do_retry:
 		list_del(&cevent->cpg_event_list);
 
 		if (is_io_request(req->rq.opcode)) {
+			int copies = sys->nr_sobjs;
+
+			if (copies > req->nr_nodes)
+				copies = req->nr_nodes;
+
+			if (__is_access_to_recoverying_objects(req)) {
+				if (req->rq.flags & SD_FLAG_CMD_DIRECT) {
+					req->rp.result = SD_RES_NEW_NODE_VER;
+					sys->nr_outstanding_io++; /* TODO: cleanup */
+					list_add_tail(&req->r_wlist, &failed_req_list);
+				} else
+					list_add_tail(&req->r_wlist, &sys->req_wait_for_obj_list);
+				continue;
+			}
 			if (__is_access_to_busy_objects(req)) {
 				list_add_tail(&req->r_wlist, &sys->req_wait_for_obj_list);
 				continue;
@@ -1328,7 +1414,10 @@ do_retry:
 
 			sys->nr_outstanding_io++;
 
-			if (req->rq.flags & SD_FLAG_CMD_DIRECT) {
+			if (is_access_local(req->entry, req->nr_vnodes,
+					    ((struct sd_obj_req *)&req->rq)->oid, copies) ||
+			    is_access_local(req->entry, req->nr_vnodes,
+					    ((struct sd_obj_req *)&req->rq)->cow_oid, copies)) {
 				int ret = check_epoch(req);
 				if (ret != SD_RES_SUCCESS) {
 					req->rp.result = ret;
@@ -1336,13 +1425,16 @@ do_retry:
 					list_add_tail(&req->r_wlist, &failed_req_list);
 					continue;
 				}
-			} else if (req->rq.opcode == SD_OP_READ_OBJ) {
+			}
+
+			if (!(req->rq.flags & SD_FLAG_CMD_DIRECT) &&
+			    req->rq.opcode == SD_OP_READ_OBJ) {
 				struct sd_obj_req *hdr = (struct sd_obj_req *)&req->rq;
 				uint32_t vdi_id = oid_to_vid(hdr->oid);
 				struct data_object_bmap *bmap;
 
 				req->check_consistency = 1;
-				if (is_data_obj(hdr->oid)) {
+				if (!is_vdi_obj(hdr->oid)) {
 					list_for_each_entry(bmap, &sys->consistent_obj_list, list) {
 						if (bmap->vdi_id == vdi_id) {
 							if (test_bit(data_oid_to_idx(hdr->oid), bmap->dobjs))
@@ -1548,19 +1640,7 @@ join_retry:
 
 	set_addr(nodeid, port);
 	sys->this_node.port = port;
-
-	ret = get_nodeid(&sys->this_node.id);
-	if (!sys->this_node.id) {
-		uint64_t hval;
-		int i;
-
-		hval = fnv_64a_buf(&sys->this_node.port,
-				   sizeof(sys->this_node.port),
-				   FNV1A_64_INIT);
-		for (i = ARRAY_SIZE(sys->this_node.addr) - 1; i >= 0; i--)
-			hval = fnv_64a_buf(&sys->this_node.addr[i], 1, hval);
-		sys->this_node.id = hval;
-	}
+	sys->this_node.nr_vnodes = SD_DEFAULT_VNODES;
 
 	if (get_latest_epoch() == 0)
 		sys->status = SD_STATUS_WAIT_FOR_FORMAT;
@@ -1580,3 +1660,20 @@ join_retry:
 	register_event(fd, group_handler, NULL);
 	return 0;
 }
+
+/* after this function is called, this node only works as a gateway */
+int leave_cluster(void)
+{
+	struct leave_message msg;
+
+	memset(&msg, 0, sizeof(msg));
+	msg.header.proto_ver = SD_SHEEP_PROTO_VER;
+	msg.header.op = SD_MSG_LEAVE;
+	msg.header.state = DM_FIN;
+	msg.header.msg_length = sizeof(msg);
+	msg.header.from = sys->this_node;
+	msg.header.nodeid = sys->this_nodeid;
+	msg.header.pid = sys->this_pid;
+
+	return send_message(sys->handle, (struct message_header *)&msg);
+}
