@@ -51,15 +51,21 @@ void resume_pending_requests(void)
 		start_cpg_event_work();
 }
 
-static int is_access_local(struct sheepdog_node_list_entry *e, int nr_nodes,
-			   uint64_t oid, int copies)
+int is_access_local(struct sheepdog_vnode_list_entry *e, int nr_nodes,
+		    uint64_t oid, int copies)
 {
 	int i, n;
 
+	if (oid == 0)
+		return 0;
+
+	if (copies > nr_nodes)
+		copies = nr_nodes;
+
 	for (i = 0; i < copies; i++) {
 		n = obj_to_sheep(e, nr_nodes, oid, i);
 
-		if (is_myself(&e[n]))
+		if (is_myself(e[n].addr, e[n].port))
 			return 1;
 	}
 
@@ -79,8 +85,10 @@ static void setup_access_to_local_objects(struct request *req)
 	copies = hdr->copies;
 	if (!copies)
 		copies = sys->nr_sobjs;
+	if (copies > req->nr_nodes)
+		copies = req->nr_nodes;
 
-	if (is_access_local(req->entry, req->nr_nodes, hdr->oid, copies))
+	if (is_access_local(req->entry, req->nr_vnodes, hdr->oid, copies))
 		req->local_oid = hdr->oid;
 }
 
@@ -89,6 +97,10 @@ static void __done(struct work *work, int idx)
 	struct request *req = container_of(work, struct request, work);
 	struct sd_req *hdr = (struct sd_req *)&req->rq;
 	int again = 0;
+	int copies = sys->nr_sobjs;
+
+	if (copies > req->nr_nodes)
+		copies = req->nr_nodes;
 
 	switch (hdr->opcode) {
 	case SD_OP_NEW_VDI:
@@ -123,7 +135,7 @@ static void __done(struct work *work, int idx)
 		     req->rp.result == SD_RES_WAIT_FOR_FORMAT)) {
 
 			req->rq.epoch = sys->epoch;
-			req->nr_nodes = setup_ordered_sd_node_list(req);
+			setup_ordered_sd_vnode_list(req);
 			setup_access_to_local_objects(req);
 
 			list_add_tail(&cevent->cpg_event_list, &sys->cpg_event_siblings);
@@ -148,6 +160,23 @@ static void __done(struct work *work, int idx)
 			bmap->vdi_id = vdi_id;
 			list_add(&bmap->list, &sys->consistent_obj_list);
 			set_bit(data_oid_to_idx(obj_hdr->oid), bmap->dobjs);
+		} else if (is_access_local(req->entry, req->nr_vnodes,
+					   ((struct sd_obj_req *)&req->rq)->oid, copies) &&
+			   req->rp.result == SD_RES_EIO) {
+			eprintf("leave from cluster\n");
+			leave_cluster();
+
+			if (req->rq.flags & SD_FLAG_CMD_DIRECT)
+				/* hack to retry */
+				req->rp.result = SD_RES_NETWORK_ERROR;
+			else {
+				req->rq.epoch = sys->epoch;
+				setup_ordered_sd_vnode_list(req);
+				setup_access_to_local_objects(req);
+
+				list_add_tail(&cevent->cpg_event_list, &sys->cpg_event_siblings);
+				again = 1;
+			}
 		}
 done:
 		resume_pending_requests();
@@ -252,8 +281,9 @@ static void queue_request(struct request *req)
 	if (!(hdr->flags & SD_FLAG_CMD_DIRECT))
 		hdr->epoch = sys->epoch;
 
-	req->nr_nodes = setup_ordered_sd_node_list(req);
-	setup_access_to_local_objects(req);
+	setup_ordered_sd_vnode_list(req);
+	if (is_io_request(hdr->opcode))
+		setup_access_to_local_objects(req);
 
 	cevent->ctype = CPG_EVENT_REQUEST;
 	list_add_tail(&cevent->cpg_event_list, &sys->cpg_event_siblings);
@@ -267,14 +297,19 @@ static struct request *alloc_request(struct client_info *ci, int data_length)
 {
 	struct request *req;
 
-	req = zalloc(sizeof(struct request) + data_length);
+	req = zalloc(sizeof(struct request));
 	if (!req)
 		return NULL;
 
 	req->ci = ci;
 	client_incref(ci);
-	if (data_length)
-		req->data = (char *)req + sizeof(*req);
+	if (data_length) {
+		req->data = valloc(data_length);
+		if (!req->data) {
+			free(req);
+			return NULL;
+		}
+	}
 
 	list_add(&req->r_siblings, &ci->reqs);
 	INIT_LIST_HEAD(&req->r_wlist);
@@ -287,6 +322,7 @@ static struct request *alloc_request(struct client_info *ci, int data_length)
 static void free_request(struct request *req)
 {
 	list_del(&req->r_siblings);
+	free(req->data);
 	free(req);
 
 	sys->nr_outstanding_reqs--;
@@ -423,11 +459,7 @@ again:
 
 	switch (ci->conn.c_tx_state) {
 	case C_IO_HEADER:
-		if (rsp->data_length)
-			ret = tx(&ci->conn, C_IO_DATA_INIT, MSG_MORE);
-		else
-			ret = tx(&ci->conn, C_IO_DATA_INIT, 0);
-
+		ret = tx(&ci->conn, C_IO_DATA_INIT, 0);
 		if (!ret)
 			break;
 
@@ -520,7 +552,7 @@ static void listen_handler(int listen_fd, int events, void *data)
 {
 	struct sockaddr_storage from;
 	socklen_t namesize;
-	int fd, ret, opt;
+	int fd, ret;
 	struct client_info *ci;
 
 	if (sys->status == SD_STATUS_SHUTDOWN) {
@@ -536,8 +568,13 @@ static void listen_handler(int listen_fd, int events, void *data)
 		return;
 	}
 
-	opt = 1;
-	ret = setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &opt, sizeof(opt));
+	ret = set_nodelay(fd);
+	if (ret) {
+		close(fd);
+		return;
+	}
+
+	ret = set_nonblocking(fd);
 	if (ret) {
 		close(fd);
 		return;
@@ -568,8 +605,8 @@ int create_listen_port(int port, void *data)
 	return create_listen_ports(port, create_listen_port_fn, data);
 }
 
-int write_object(struct sheepdog_node_list_entry *e,
-		 int nodes, uint32_t node_version,
+int write_object(struct sheepdog_vnode_list_entry *e,
+		 int vnodes, int nodes, uint32_t node_version,
 		 uint64_t oid, char *data, unsigned int datalen,
 		 uint64_t offset, int nr, int create)
 {
@@ -583,7 +620,7 @@ int write_object(struct sheepdog_node_list_entry *e,
 	for (i = 0; i < nr; i++) {
 		unsigned rlen = 0, wlen = datalen;
 
-		n = obj_to_sheep(e, nodes, oid, i);
+		n = obj_to_sheep(e, vnodes, oid, i);
 
 		if (memcmp(e[n].addr, sys->this_node.addr, sizeof(e[n].addr)) == 0 &&
 		    e[n].port == sys->this_node.port) {
@@ -631,8 +668,8 @@ int write_object(struct sheepdog_node_list_entry *e,
 	return !success;
 }
 
-int read_object(struct sheepdog_node_list_entry *e,
-		int nodes, uint32_t node_version,
+int read_object(struct sheepdog_vnode_list_entry *e,
+		int vnodes, int nodes, uint32_t node_version,
 		uint64_t oid, char *data, unsigned int datalen,
 		uint64_t offset, int nr)
 {
@@ -646,7 +683,7 @@ int read_object(struct sheepdog_node_list_entry *e,
 
 	/* search a local object first */
 	for (i = 0; i < nr; i++) {
-		n = obj_to_sheep(e, nodes, oid, i);
+		n = obj_to_sheep(e, vnodes, oid, i);
 
 		if (memcmp(e[n].addr, sys->this_node.addr, sizeof(e[n].addr)) == 0 &&
 		    e[n].port == sys->this_node.port) {
@@ -666,7 +703,7 @@ int read_object(struct sheepdog_node_list_entry *e,
 	for (i = 0; i < nr; i++) {
 		unsigned wlen = 0, rlen = datalen;
 
-		n = obj_to_sheep(e, nodes, oid, i);
+		n = obj_to_sheep(e, vnodes, oid, i);
 
 		addr_to_str(name, sizeof(name), e[n].addr, 0);
 
@@ -703,8 +740,8 @@ int read_object(struct sheepdog_node_list_entry *e,
 	return -last_error;
 }
 
-int remove_object(struct sheepdog_node_list_entry *e,
-		  int nodes, uint32_t node_version,
+int remove_object(struct sheepdog_vnode_list_entry *e,
+		  int vnodes, int nodes, uint32_t node_version,
 		  uint64_t oid, int nr)
 {
 	char name[128];
@@ -718,7 +755,7 @@ int remove_object(struct sheepdog_node_list_entry *e,
 	for (i = 0; i < nr; i++) {
 		unsigned wlen = 0, rlen = 0;
 
-		n = obj_to_sheep(e, nodes, oid, i);
+		n = obj_to_sheep(e, vnodes, oid, i);
 
 		addr_to_str(name, sizeof(name), e[n].addr, 0);
 
@@ -749,33 +786,7 @@ int remove_object(struct sheepdog_node_list_entry *e,
 	return 0;
 }
 
-static int set_nonblocking(int fd)
-{
-	int ret;
-
-	ret = fcntl(fd, F_GETFL);
-	if (ret < 0) {
-		eprintf("can't fcntl (F_GETFL), %m\n");
-		close(fd);
-	} else {
-		ret = fcntl(fd, F_SETFL, ret | O_NONBLOCK);
-		if (ret < 0)
-			eprintf("can't fcntl (O_NONBLOCK), %m\n");
-	}
-
-	return ret;
-}
-
-static int set_nodelay(int fd)
-{
-	int ret, opt;
-
-	opt = 1;
-	ret = setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &opt, sizeof(opt));
-	return ret;
-}
-
-int get_sheep_fd(struct sheepdog_node_list_entry *e, int node_idx,
+int get_sheep_fd(uint8_t *addr, uint16_t port, int node_idx,
 		 uint32_t epoch, int worker_idx)
 {
 	static int cached_fds[NR_WORKER_THREAD][SD_MAX_NODES];
@@ -817,19 +828,12 @@ int get_sheep_fd(struct sheepdog_node_list_entry *e, int node_idx,
 		return fd;
 	}
 
-	addr_to_str(name, sizeof(name), e[node_idx].addr, 0);
+	addr_to_str(name, sizeof(name), addr, 0);
 
-	fd = connect_to(name, e[node_idx].port);
+	fd = connect_to(name, port);
 	if (fd < 0)
 		return -1;
 
-	ret = set_nonblocking(fd);
-	if (ret) {
-		eprintf("%m\n");
-		close(fd);
-		return -1;
-	}
-
 	ret = set_nodelay(fd);
 	if (ret) {
 		eprintf("%m\n");
