@@ -16,6 +16,7 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <time.h>
+#include <signal.h>
 #include <linux/limits.h>
 #include <sys/syslog.h>
 
@@ -32,11 +33,12 @@ static struct option const long_options[] = {
 	{"foreground", no_argument, NULL, 'f'},
 	{"loglevel", required_argument, NULL, 'l'},
 	{"debug", no_argument, NULL, 'd'},
+	{"directio", no_argument, NULL, 'D'},
 	{"help", no_argument, NULL, 'h'},
 	{NULL, 0, NULL, 0},
 };
 
-static const char *short_options = "p:fl:dh";
+static const char *short_options = "p:fl:dDh";
 
 static void usage(int status)
 {
@@ -51,6 +53,7 @@ Sheepdog Daemon, version %s\n\
   -f, --foreground        make the program run in the foreground\n\
   -l, --loglevel          specify the message level printed by default\n\
   -d, --debug             print debug messages\n\
+  -D, --directio          use direct IO\n\
   -h, --help              display this help and exit\n\
 ", PACKAGE_VERSION);
 	}
@@ -69,6 +72,8 @@ int main(int argc, char **argv)
 	int log_level = LOG_INFO;
 	char path[PATH_MAX];
 
+	signal(SIGPIPE, SIG_IGN);
+
 	while ((ch = getopt_long(argc, argv, short_options, long_options,
 				 &longindex)) >= 0) {
 		switch (ch) {
@@ -85,6 +90,10 @@ int main(int argc, char **argv)
 			/* removed soon. use loglevel instead */
 			log_level = LOG_DEBUG;
 			break;
+		case 'D':
+			dprintf("direct IO mode\n");
+			sys->use_directio = 1;
+			break;
 		case 'h':
 			usage(0);
 			break;
