@@ -26,7 +26,6 @@
 
 #define ANAME_CTIME "user.sheepdog.ctime"
 #define ANAME_COPIES "user.sheepdog.copies"
-#define ANAME_NODEID "user.sheepdog.nodeid"
 
 static char *obj_path;
 static char *epoch_path;
@@ -102,22 +101,12 @@ static int stat_sheep(uint64_t *store_size, uint64_t *store_free, uint32_t epoch
 
 	closedir(dir);
 
-	*store_size = vs.f_frsize * vs.f_bfree + used;
-	*store_free = vs.f_frsize * vs.f_bfree;
+	*store_size = (uint64_t)vs.f_frsize * vs.f_bfree + used;
+	*store_free = (uint64_t)vs.f_frsize * vs.f_bfree;
 
 	return SD_RES_SUCCESS;
 }
 
-static int is_obj_in_range(uint64_t oid, uint64_t start, uint64_t end)
-{
-	uint64_t hval = fnv_64a_buf(&oid, sizeof(oid), FNV1A_64_INIT);
-
-	if (start < end)
-		return (start < hval && hval <= end);
-	else
-		return (start < hval || hval <= end);
-}
-
 static int get_obj_list(struct request *req)
 {
 	DIR *dir;
@@ -125,17 +114,12 @@ static int get_obj_list(struct request *req)
 	struct sd_list_req *hdr = (struct sd_list_req *)&req->rq;
 	struct sd_list_rsp *rsp = (struct sd_list_rsp *)&req->rp;
 	uint64_t oid;
-	uint64_t start_hash = hdr->start;
-	uint64_t end_hash = hdr->end;
 	uint32_t epoch = hdr->tgt_epoch;
 	char path[1024];
 	uint64_t *p = (uint64_t *)req->data;
 	int nr = 0;
 	uint64_t *objlist = NULL;
 	int obj_nr = 0, fd, i;
-	struct sheepdog_node_list_entry *e;
-	int e_nr;
-	int idx;
 	int res = SD_RES_SUCCESS;
 	int buf_len;
 	char *buf;
@@ -166,11 +150,7 @@ static int get_obj_list(struct request *req)
 	obj_nr /= sizeof(uint64_t);
 	objlist = (uint64_t *)buf;
 	for (i = 0; i < obj_nr; i++) {
-		if (is_obj_in_range(objlist[i], start_hash, end_hash)) {
-			dprintf("%u, %016"PRIx64", %016"PRIx64" %016"PRIx64"\n", epoch,
-				objlist[i], start_hash, end_hash);
-			p[nr++] = objlist[i];
-		}
+		p[nr++] = objlist[i];
 
 		if (nr * sizeof(uint64_t) >= hdr->data_length)
 			break;
@@ -203,11 +183,7 @@ local:
 		if (i < obj_nr)
 			continue;
 
-		if (is_obj_in_range(oid, start_hash, end_hash)) {
-			dprintf("%u, %016"PRIx64", %016"PRIx64" %016"PRIx64"\n", epoch,
-				oid, start_hash, end_hash);
-			p[nr++] = oid;
-		}
+		p[nr++] = oid;
 
 		if (nr * sizeof(uint64_t) >= hdr->data_length)
 			break;
@@ -215,39 +191,7 @@ local:
 
 	eprintf("nr = %"PRIu32"\n", nr);
 
-	e_nr = epoch_log_read(epoch, buf, buf_len);
-	e_nr /= sizeof(*e);
-	e = (struct sheepdog_node_list_entry *)buf;
-
-	if (e_nr <= sys->nr_sobjs) {
-		rsp->next = end_hash;
-		closedir(dir);
-		goto out;
-	}
-
-	for (idx = 0; idx < e_nr; idx++) {
-		if (e[idx].id == sys->this_node.id)
-			break;
-	}
-	if (idx != e_nr) {
-		uint64_t hval = e[idx % e_nr].id;
-
-		rsp->next = end_hash;
-
-		if (start_hash < end_hash) {
-			if (start_hash < hval && hval <= end_hash)
-				rsp->next = hval;
-		} else
-			if (start_hash < hval || hval <= end_hash)
-				rsp->next = hval;
-
-		dprintf("%u, %016"PRIx64", %016"PRIx64" %016"PRIx64"\n", epoch, hval,
-			start_hash, end_hash);
-	} else
-		res = SD_RES_SYSTEM_ERROR;
-
 	closedir(dir);
-
 out:
 	free(buf);
 	rsp->data_length = nr * sizeof(uint64_t);
@@ -265,19 +209,19 @@ static int read_from_one(struct request *req, uint32_t epoch, uint64_t oid,
 	int i, n, nr, fd, ret;
 	unsigned wlen, rlen;
 	char name[128];
-	struct sheepdog_node_list_entry *e;
+	struct sheepdog_vnode_list_entry *e;
 	struct sd_obj_req hdr;
 	struct sd_obj_rsp *rsp = (struct sd_obj_rsp *)&hdr;
 
 	e = req->entry;
-	nr = req->nr_nodes;
+	nr = req->nr_vnodes;
 
 	for (i = 0; i < nr; i++) {
 		n = obj_to_sheep(e, nr, oid, i);
 
 		addr_to_str(name, sizeof(name), e[n].addr, 0);
 
-		if (is_myself(&e[n])) {
+		if (is_myself(e[n].addr, e[n].port)) {
 			fd = ob_open(epoch, oid, 0, &ret);
 			if (fd < 0 || ret != 0)
 				continue;
@@ -352,12 +296,12 @@ static int forward_read_obj_req(struct request *req, int idx)
 	unsigned wlen, rlen;
 	struct sd_obj_req hdr = *(struct sd_obj_req *)&req->rq;
 	struct sd_obj_rsp *rsp = (struct sd_obj_rsp *)&hdr;
-	struct sheepdog_node_list_entry *e;
+	struct sheepdog_vnode_list_entry *e;
 	uint64_t oid = hdr.oid;
 	int copies;
 
 	e = req->entry;
-	nr = req->nr_nodes;
+	nr = req->nr_vnodes;
 
 	copies = hdr.copies;
 
@@ -373,7 +317,7 @@ static int forward_read_obj_req(struct request *req, int idx)
 	for (i = 0; i < copies; i++) {
 		n = obj_to_sheep(e, nr, oid, i);
 
-		if (is_myself(&e[n])) {
+		if (is_myself(e[n].addr, e[n].port)) {
 			ret = store_queue_request_local(req, hdr.epoch);
 			goto out;
 		}
@@ -381,7 +325,7 @@ static int forward_read_obj_req(struct request *req, int idx)
 
 	n = obj_to_sheep(e, nr, oid, 0);
 
-	fd = get_sheep_fd(e, n, hdr.epoch, idx);
+	fd = get_sheep_fd(e[n].addr, e[n].port, e[n].node_idx, hdr.epoch, idx);
 	if (fd < 0) {
 		ret = SD_RES_NETWORK_ERROR;
 		goto out;
@@ -409,15 +353,15 @@ static int forward_write_obj_req(struct request *req, int idx)
 	char name[128];
 	struct sd_obj_req hdr = *(struct sd_obj_req *)&req->rq;
 	struct sd_obj_rsp *rsp = (struct sd_obj_rsp *)&req->rp;
-	struct sheepdog_node_list_entry *e;
+	struct sheepdog_vnode_list_entry *e;
 	uint64_t oid = hdr.oid;
 	int copies;
 	struct pollfd pfds[SD_MAX_REDUNDANCY];
-	int done, nr_fds, local = 0;
+	int nr_fds, local = 0;
 
 	dprintf("%"PRIx64"\n", oid);
 	e = req->entry;
-	nr = req->nr_nodes;
+	nr = req->nr_vnodes;
 
 	copies = hdr.copies;
 
@@ -428,7 +372,6 @@ static int forward_write_obj_req(struct request *req, int idx)
 		copies = req->nr_nodes;
 
 	nr_fds = 0;
-	done = 0;
 	memset(pfds, 0, sizeof(pfds));
 	for (i = 0; i < ARRAY_SIZE(pfds); i++)
 		pfds[i].fd = -1;
@@ -443,12 +386,12 @@ static int forward_write_obj_req(struct request *req, int idx)
 
 		addr_to_str(name, sizeof(name), e[n].addr, 0);
 
-		if (is_myself(&e[n])) {
+		if (is_myself(e[n].addr, e[n].port)) {
 			local = 1;
 			continue;
 		}
 
-		fd = get_sheep_fd(e, n, hdr.epoch, idx);
+		fd = get_sheep_fd(e[n].addr, e[n].port, e[n].node_idx, hdr.epoch, idx);
 		if (fd < 0) {
 			eprintf("failed to connect to %s:%"PRIu32"\n", name, e[n].port);
 			ret = SD_RES_NETWORK_ERROR;
@@ -482,52 +425,50 @@ static int forward_write_obj_req(struct request *req, int idx)
 		}
 	}
 
+	ret = SD_RES_SUCCESS;
 again:
-	ret = poll(pfds, nr_fds, -1);
-
-	if (ret < 0) {
+	if (poll(pfds, nr_fds, -1) < 0) {
 		if (errno == EINTR)
 			goto again;
 
 		ret = SD_RES_EIO;
-		goto out;
 	}
 
 	for (i = 0; i < nr_fds; i++) {
 		if (pfds[i].fd < 0)
-			continue;
+			break;
 
-		if (pfds[i].revents & POLLERR || pfds[i].revents & POLLHUP) {
-			ret = SD_RES_EIO;
-			goto out;
+		if (pfds[i].revents & POLLERR || pfds[i].revents & POLLHUP || pfds[i].revents & POLLNVAL) {
+			ret = SD_RES_NETWORK_ERROR;
+			break;
 		}
 
 		if (!(pfds[i].revents & POLLIN))
 			continue;
 
-		ret = do_read(pfds[i].fd, rsp, sizeof(*rsp));
-
-		if (ret) {
+		if (do_read(pfds[i].fd, rsp, sizeof(*rsp))) {
 			eprintf("failed to get a rsp, %m\n");
 			ret = SD_RES_NETWORK_ERROR;
-			goto out;
+			break;
 		}
 
 		if (rsp->result != SD_RES_SUCCESS) {
 			eprintf("fail %"PRIu32"\n", rsp->result);
 			ret = rsp->result;
-			goto out;
 		}
 
-		done++;
+		break;
+	}
+	if (i < nr_fds) {
+		nr_fds--;
+		memmove(pfds + i, pfds + i + 1, sizeof(*pfds) * (nr_fds - i));
 	}
 
-	dprintf("%"PRIx64" %"PRIu32" %"PRIu32"\n", oid, nr_fds, done);
+	dprintf("%"PRIx64" %"PRIu32"\n", oid, nr_fds);
 
-	if (done != nr_fds)
+	if (nr_fds > 0) {
 		goto again;
-
-	ret = SD_RES_SUCCESS;
+	}
 out:
 	return ret;
 }
@@ -535,17 +476,29 @@ out:
 static int ob_open(uint32_t epoch, uint64_t oid, int aflags, int *ret)
 {
 	char path[1024];
-	int flags = O_SYNC | O_RDWR | aflags;
+	int flags;
 	int fd;
 
+	if (sys->use_directio && is_data_obj(oid))
+		flags = O_DIRECT | O_RDWR | aflags;
+	else
+		flags = O_SYNC | O_RDWR | aflags;
+
 	snprintf(path, sizeof(path), "%s%08u/%016" PRIx64, obj_path, epoch, oid);
 
 	fd = open(path, flags, def_fmode);
 	if (fd < 0) {
 		eprintf("failed to open %s, %s\n", path, strerror(errno));
-		if (errno == ENOENT)
+		if (errno == ENOENT) {
+			struct stat s;
+
 			*ret = SD_RES_NO_OBJ;
-		else
+			if (stat(obj_path, &s) < 0) {
+				/* store directory is corrupted */
+				eprintf("corrupted\n");
+				*ret = SD_RES_EIO;
+			}
+		} else
 			*ret = SD_RES_UNKNOWN;
 	} else
 		*ret = 0;
@@ -664,7 +617,7 @@ static int store_queue_request_local(struct request *req, uint32_t epoch)
 		if (hdr->flags & SD_FLAG_CMD_COW) {
 			dprintf("%" PRIu64 ", %" PRIx64 "\n", oid, hdr->cow_oid);
 
-			buf = zalloc(SD_DATA_OBJ_SIZE);
+			buf = valloc(SD_DATA_OBJ_SIZE);
 			if (!buf) {
 				eprintf("failed to allocate memory\n");
 				ret = SD_RES_NO_MEM;
@@ -688,10 +641,19 @@ static int store_queue_request_local(struct request *req, uint32_t epoch)
 			free(buf);
 			buf = NULL;
 		} else {
-			int zero = 0;
+			int size = SECTOR_SIZE;
+			buf = valloc(size);
+			if (!buf) {
+				eprintf("failed to allocate memory\n");
+				ret = SD_RES_NO_MEM;
+				goto out;
+			}
+			memset(buf, 0, size);
+			ret = pwrite64(fd, buf, size, SD_DATA_OBJ_SIZE - size);
+			free(buf);
+			buf = NULL;
 
-			ret = pwrite64(fd, &zero, sizeof(zero), SD_DATA_OBJ_SIZE - sizeof(zero));
-			if (ret != sizeof(zero)) {
+			if (ret != size) {
 				if (errno == ENOSPC)
 					ret = SD_RES_NO_SPACE;
 				else
@@ -744,7 +706,7 @@ static int store_queue_request_local(struct request *req, uint32_t epoch)
 		if (hdr->flags & SD_FLAG_CMD_TRUNCATE)
 			ftruncate(fd, hdr->offset + hdr->data_length);
 
-		if (!is_data_obj(oid)) {
+		if (is_vdi_obj(oid)) {
 			jd.jdf_epoch = epoch;
 			jd.jdf_oid = oid;
 			jd.jdf_target_fd = fd;
@@ -802,16 +764,19 @@ static int fix_object_consistency(struct request *req, int idx)
 	void *data = req->data, *buf;
 	uint64_t oid = hdr->oid;
 
-	if (is_data_obj(hdr->oid))
-		data_length = SD_DATA_OBJ_SIZE;
-	else
+	if (is_vdi_obj(hdr->oid))
 		data_length = sizeof(struct sheepdog_inode);
+	else if (is_vdi_attr_obj(hdr->oid))
+		data_length = SD_MAX_VDI_ATTR_VALUE_LEN;
+	else
+		data_length = SD_DATA_OBJ_SIZE;
 
-	buf = zalloc(data_length);
+	buf = valloc(data_length);
 	if (buf == NULL) {
 		eprintf("out of memory\n");
 		goto out;
 	}
+	memset(buf, 0, data_length);
 
 	req->data = buf;
 	hdr->offset = 0;
@@ -819,7 +784,7 @@ static int fix_object_consistency(struct request *req, int idx)
 	hdr->opcode = SD_OP_READ_OBJ;
 	hdr->flags = 0;
 	ret = forward_read_obj_req(req, idx);
-	if (ret < 0) {
+	if (ret != SD_RES_SUCCESS) {
 		eprintf("failed to read object, %d\n", ret);
 		goto out;
 	}
@@ -828,7 +793,7 @@ static int fix_object_consistency(struct request *req, int idx)
 	hdr->flags = SD_FLAG_CMD_WRITE;
 	hdr->oid = oid;
 	ret = forward_write_obj_req(req, idx);
-	if (ret < 0) {
+	if (ret != SD_RES_SUCCESS) {
 		eprintf("failed to write object, %d\n", ret);
 		goto out;
 	}
@@ -871,7 +836,7 @@ void store_queue_request(struct work *work, int idx)
 		/* fix object consistency when we read the object for the first time */
 		if (req->check_consistency) {
 			ret = fix_object_consistency(req, idx);
-			if (ret < 0)
+			if (ret != SD_RES_SUCCESS)
 				goto out;
 		}
 
@@ -887,8 +852,10 @@ out:
 	if (ret != SD_RES_SUCCESS) {
 		dprintf("failed, %"PRIu32", %x, %" PRIx64" , %u, %"PRIu32"\n",
 			idx, opcode, oid, epoch, ret);
-		rsp->result = ret;
+		if (!(ret == SD_RES_NO_OBJ && hdr->flags & SD_FLAG_CMD_RECOVERY))
+			rsp->data_length = 0;
 	}
+	rsp->result = ret;
 }
 
 int epoch_log_write(uint32_t epoch, char *buf, int len)
@@ -1053,24 +1020,36 @@ uint64_t get_cluster_ctime(void)
 	return ctime;
 }
 
-static int node_distance(int my, int her, int nr)
-{
-	return (my + nr - her) % nr;
-}
-
-static int contains_node(uint64_t id, struct sheepdog_node_list_entry *entry,
+/*
+ * contains_node - checks that the node id is included in the target nodes
+ *
+ * The target nodes to store replicated objects are the first N nodes
+ * from the base_idx'th on the consistent hash ring, where N is the
+ * number of copies of objects.
+ */
+static int contains_node(struct sheepdog_vnode_list_entry *key,
+			 struct sheepdog_vnode_list_entry *entry,
 			 int nr, int base_idx)
 {
 	int i;
 
 	for (i = 0; i < sys->nr_sobjs; i++) {
-		if (entry[(base_idx + i) % nr].id == id)
-			return (base_idx + i) % nr;
+		int idx = get_nth_node(entry, nr, base_idx, i);
+		if (memcmp(key->addr, entry[idx].addr, sizeof(key->addr)) == 0
+		    && key->port == entry[idx].port)
+			return idx;
 	}
 	return -1;
 }
 
+enum rw_state {
+	RW_INIT,
+	RW_RUN,
+};
+
 struct recovery_work {
+	enum rw_state state;
+
 	uint32_t epoch;
 	uint32_t done;
 
@@ -1079,24 +1058,57 @@ struct recovery_work {
 	struct work work;
 	struct list_head rw_siblings;
 
+	int nr_blocking;
 	int count;
-	char *buf;
+	uint64_t *oids;
 };
 
 static LIST_HEAD(recovery_work_list);
 static struct recovery_work *recovering_work;
-static uint64_t blocking_oid;
 
-static int find_tgt_node(struct sheepdog_node_list_entry *old_entry, int old_nr, int old_idx,
-			 struct sheepdog_node_list_entry *cur_entry, int cur_nr, int cur_idx,
+/*
+ * find_tgt_node - find the node from which we should recover objects
+ *
+ * This function compares two node lists, the current target nodes and
+ * the previous target nodes, and finds the node from the previous
+ * target nodes which corresponds to the copy_idx'th node of the
+ * current target nodes.  The correspondence is injective and
+ * maximizes the number of nodes which can recover objects locally.
+ *
+ * For example, consider the number of redundancy is 5, the consistent
+ * hash ring is {A, B, C, D, E, F}, and the node G is newly added.
+ * The parameters of this function are
+ *   old_entry = {A, B, C, D, E, F},    old_nr = 6, old_idx = 3
+ *   cur_entry = {A, B, C, D, E, F, G}, cur_nr = 7, cur_idx = 3
+ *
+ * In this case:
+ *   the previous target nodes: {D, E, F, A, B}
+ *     (the first 5 nodes from the 3rd node on the previous hash ring)
+ *   the current target nodes : {D, E, F, G, A}
+ *     (the first 5 nodes from the 3rd node on the current hash ring)
+ *
+ * The correspondence between copy_idx and return value are as follows:
+ * ----------------------------
+ * copy_idx       0  1  2  3  4
+ * src_node       D  E  F  G  A
+ * tgt_node       D  E  F  B  A
+ * return value   0  1  2  4  3
+ * ----------------------------
+ *
+ * The node D, E, F, and A can recover objects from local, and the
+ * node G recovers from the node B.
+ */
+static int find_tgt_node(struct sheepdog_vnode_list_entry *old_entry, int old_nr, int old_idx,
+			 struct sheepdog_vnode_list_entry *cur_entry, int cur_nr, int cur_idx,
 			 int copy_idx)
 {
 	int i, j, idx;
 
 	dprintf("%"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32"\n", old_idx, old_nr, cur_idx, cur_nr, copy_idx);
 
-	if (copy_idx < cur_nr) {
-		idx = contains_node(cur_entry[(cur_idx + copy_idx) % cur_nr].id,
+	if (copy_idx < sys->nr_sobjs) {
+		/* If the same node is in the previous target nodes, return its index */
+		idx = contains_node(cur_entry + get_nth_node(cur_entry, cur_nr, cur_idx, copy_idx),
 				    old_entry, old_nr, old_idx);
 		if (idx >= 0) {
 			dprintf("%"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32"\n", idx, copy_idx, cur_idx, cur_nr);
@@ -1105,48 +1117,55 @@ static int find_tgt_node(struct sheepdog_node_list_entry *old_entry, int old_nr,
 	}
 
 	for (i = 0, j = 0; ; i++, j++) {
-		if (i < cur_nr) {
-			idx = contains_node(cur_entry[(cur_idx + i) % cur_nr].id,
+		if (i < copy_idx) {
+			/* Skip if the node can recover from its local */
+			idx = contains_node(cur_entry + get_nth_node(cur_entry, cur_nr, cur_idx, i),
 					    old_entry, old_nr, old_idx);
 			if (idx >= 0)
 				continue;
 
-			while (contains_node(old_entry[(old_idx + j) % old_nr].id,
-					     cur_entry, cur_nr, cur_idx) >= 0 && j < old_nr)
+			/* Find the next target which needs to recover from remote */
+			while (j < sys->nr_sobjs &&
+			       contains_node(old_entry + get_nth_node(old_entry, old_nr, old_idx, j),
+					     cur_entry, cur_nr, cur_idx) >= 0)
 				j++;
 		}
-		if (j == old_nr) {
-			/* old_nr should be smaller than sys->nr_sobjs */
-			if (old_nr >= sys->nr_sobjs)
-				eprintf("bug: %"PRIu32", %"PRIu32"\n", old_nr, sys->nr_sobjs);
-
+		if (j == sys->nr_sobjs) {
+			/*
+			 * Cannot find the target because the number of nodes
+			 * is smaller than the number of copies.  We can select
+			 * any node in this case, so select the first one.
+			 */
 			return old_idx;
 		}
 
 		if (i == copy_idx) {
-			dprintf("%"PRIu32", %"PRIu32", %"PRIu32"\n", (old_idx + j) % old_nr, copy_idx,
-				(cur_idx + i) % cur_nr);
-			return (old_idx + j) % old_nr;
+			/* Found the target node correspoinding to copy_idx */
+			dprintf("%"PRIu32", %"PRIu32", %"PRIu32"\n",
+				get_nth_node(old_entry, old_nr, old_idx, j),
+				copy_idx, (cur_idx + i) % cur_nr);
+			return get_nth_node(old_entry, old_nr, old_idx, j);
 		}
 
 	}
+
 	return -1;
 }
 
 static int __recover_one(struct recovery_work *rw,
-			 struct sheepdog_node_list_entry *_old_entry, int old_nr,
-			 struct sheepdog_node_list_entry *_cur_entry, int cur_nr, int cur_idx,
+			 struct sheepdog_vnode_list_entry *_old_entry, int old_nr,
+			 struct sheepdog_vnode_list_entry *_cur_entry, int cur_nr, int cur_idx,
 			 int copy_idx, uint32_t epoch, uint32_t tgt_epoch,
 			 uint64_t oid, char *buf, int buf_len)
 {
-	struct sheepdog_node_list_entry *e;
+	struct sheepdog_vnode_list_entry *e;
 	struct sd_obj_req hdr;
 	struct sd_obj_rsp *rsp = (struct sd_obj_rsp *)&hdr;
 	char name[128];
 	unsigned wlen = 0, rlen;
 	int fd, ret;
-	struct sheepdog_node_list_entry old_entry[SD_MAX_NODES],
-		cur_entry[SD_MAX_NODES], *next_entry;
+	struct sheepdog_vnode_list_entry old_entry[SD_MAX_VNODES],
+		cur_entry[SD_MAX_VNODES], next_entry[SD_MAX_VNODES];
 	int next_nr;
 	int tgt_idx = -1;
 	int old_idx;
@@ -1164,7 +1183,7 @@ next:
 	}
 	e = old_entry + tgt_idx;
 
-	if (e->id == sys->this_node.id) {
+	if (is_myself(e->addr, e->port)) {
 		char old[PATH_MAX], new[PATH_MAX];
 
 		snprintf(old, sizeof(old), "%s%08u/%016" PRIx64, obj_path,
@@ -1176,13 +1195,14 @@ next:
 			return 0;
 
 		if (errno == ENOENT) {
-			next_nr = epoch_log_read(tgt_epoch, buf, buf_len);
+			next_nr = epoch_log_read(tgt_epoch - 1, buf, buf_len);
 			if (next_nr <= 0) {
-				eprintf("no previous epoch, %"PRIu32"\n", tgt_epoch);
+				eprintf("no previous epoch, %"PRIu32"\n", tgt_epoch - 1);
 				return -1;
 			}
-			next_entry = (struct sheepdog_node_list_entry *)buf;
-			next_nr /= sizeof(*next_entry);
+			next_nr = nodes_to_vnodes((struct sheepdog_node_list_entry *)buf,
+						  next_nr / sizeof(struct sheepdog_node_list_entry),
+						  next_entry);
 			goto not_found;
 		}
 
@@ -1198,10 +1218,12 @@ next:
 		return -1;
 	}
 
-	if (is_data_obj(oid))
-		rlen = SD_DATA_OBJ_SIZE;
-	else
+	if (is_vdi_obj(oid))
 		rlen = sizeof(struct sheepdog_inode);
+	else if (is_vdi_attr_obj(oid))
+		rlen = SD_MAX_VDI_ATTR_VALUE_LEN;
+	else
+		rlen = SD_DATA_OBJ_SIZE;
 
 	memset(&hdr, 0, sizeof(hdr));
 	hdr.opcode = SD_OP_READ_OBJ;
@@ -1262,7 +1284,8 @@ next:
 		return 0;
 	}
 
-	if (rsp->result == SD_RES_NEW_NODE_VER || rsp->result == SD_RES_OLD_NODE_VER) {
+	if (rsp->result == SD_RES_NEW_NODE_VER || rsp->result == SD_RES_OLD_NODE_VER
+	    || rsp->result == SD_RES_NETWORK_ERROR) {
 		eprintf("try again, %"PRIu32", %"PRIx64"\n", rsp->result, oid);
 		rw->retry = 1;
 		return 0;
@@ -1272,14 +1295,21 @@ next:
 		eprintf("%"PRIu32"\n", rsp->result);
 		return -1;
 	}
-	next_entry = (struct sheepdog_node_list_entry *)buf;
-	next_nr = rsp->data_length / sizeof(*old_entry);
+	next_nr = nodes_to_vnodes((struct sheepdog_node_list_entry *)buf,
+				  rsp->data_length / sizeof(struct sheepdog_node_list_entry),
+				  next_entry);
 
 not_found:
-	copy_idx = node_distance(tgt_idx, old_idx, old_nr);
+	for (copy_idx = 0; copy_idx < sys->nr_sobjs; copy_idx++)
+		if (get_nth_node(old_entry, old_nr, old_idx, copy_idx) == tgt_idx)
+			break;
+	if (copy_idx == sys->nr_sobjs) {
+		eprintf("bug: cannot find the proper copy_idx\n");
+		return -1;
+	}
+
 	dprintf("%"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32"\n", rsp->result, rsp->data_length, tgt_idx,
 		old_idx, old_nr, copy_idx);
-
 	memcpy(cur_entry, old_entry, sizeof(*old_entry) * old_nr);
 	cur_nr = old_nr;
 	cur_idx = old_idx;
@@ -1296,19 +1326,19 @@ static void recover_one(struct work *work, int idx)
 	struct recovery_work *rw = container_of(work, struct recovery_work, work);
 	char *buf = NULL;
 	int ret;
-	uint64_t oid = *(((uint64_t *)rw->buf) + rw->done);
-	struct sheepdog_node_list_entry old_entry[SD_MAX_NODES],
-		cur_entry[SD_MAX_NODES];
-	int old_nr, cur_nr;
+	uint64_t oid = rw->oids[rw->done];
+	struct sheepdog_node_list_entry old_nodes[SD_MAX_NODES];
+	struct sheepdog_node_list_entry cur_nodes[SD_MAX_NODES];
+	struct sheepdog_vnode_list_entry old_vnodes[SD_MAX_VNODES];
+	struct sheepdog_vnode_list_entry cur_vnodes[SD_MAX_VNODES];
+	int old_nr_nodes, cur_nr_nodes, old_nr_vnodes, cur_nr_vnodes;
 	uint32_t epoch = rw->epoch;
-	int i, my_idx = -1, copy_idx = 0, cur_idx = -1;
+	int i, copy_idx = 0, cur_idx = -1;
 	int fd;
+	int nr_objs;
 
 	eprintf("%"PRIu32" %"PRIu32", %16"PRIx64"\n", rw->done, rw->count, oid);
 
-	if (blocking_oid)
-		oid = blocking_oid; /* recover the blocked object first */
-
 	fd = ob_open(epoch, oid, 0, &ret);
 	if (fd != -1) {
 		/* the object is already recovered */
@@ -1316,40 +1346,57 @@ static void recover_one(struct work *work, int idx)
 		goto out;
 	}
 
-	if (is_data_obj(oid))
-		buf = malloc(SD_DATA_OBJ_SIZE);
-	else
+	if (is_vdi_obj(oid))
 		buf = malloc(sizeof(struct sheepdog_inode));
+	else if (is_vdi_attr_obj(oid))
+		buf = malloc(SD_MAX_VDI_ATTR_VALUE_LEN);
+	else if (is_data_obj(oid))
+		buf = valloc(SD_DATA_OBJ_SIZE);
+	else
+		buf = malloc(SD_DATA_OBJ_SIZE);
 
-	cur_nr = epoch_log_read(epoch, (char *)cur_entry, sizeof(cur_entry));
-	if (cur_nr <= 0) {
+	cur_nr_nodes = epoch_log_read(epoch, (char *)cur_nodes, sizeof(cur_nodes));
+	if (cur_nr_nodes <= 0) {
 		eprintf("failed to read current epoch, %"PRIu32"\n", epoch);
 		goto out;
 	}
-	cur_nr /= sizeof(struct sheepdog_node_list_entry);
+	cur_nr_nodes /= sizeof(struct sheepdog_node_list_entry);
 
-	old_nr = epoch_log_read(epoch - 1, (char *)old_entry, sizeof(old_entry));
-	if (old_nr <= 0) {
+	old_nr_nodes = epoch_log_read(epoch - 1, (char *)old_nodes, sizeof(old_nodes));
+	if (old_nr_nodes <= 0) {
 		eprintf("failed to read previous epoch, %"PRIu32"\n", epoch - 1);
 		goto fail;
 	}
-	old_nr /= sizeof(struct sheepdog_node_list_entry);
+	old_nr_nodes /= sizeof(struct sheepdog_node_list_entry);
+
+	old_nr_vnodes = nodes_to_vnodes(old_nodes, old_nr_nodes, old_vnodes);
+	cur_nr_vnodes = nodes_to_vnodes(cur_nodes, cur_nr_nodes, cur_vnodes);
 
 	if (!sys->nr_sobjs)
 		goto fail;
 
-	cur_idx = obj_to_sheep(cur_entry, cur_nr, oid, 0);
+	cur_idx = obj_to_sheep(cur_vnodes, cur_nr_vnodes, oid, 0);
 
-	for (i = 0; i < cur_nr; i++) {
-		if (cur_entry[i].id == sys->this_node.id) {
-			my_idx = i;
+	nr_objs = sys->nr_sobjs;
+	if (nr_objs > cur_nr_nodes)
+		nr_objs = cur_nr_nodes;
+
+	copy_idx = -1;
+	for (i = 0; i < nr_objs; i++) {
+		int n = obj_to_sheep(cur_vnodes, cur_nr_vnodes, oid, i);
+		if (is_myself(cur_vnodes[n].addr, cur_vnodes[n].port)) {
+			copy_idx = i;
 			break;
 		}
 	}
-	copy_idx = node_distance(my_idx, cur_idx, cur_nr);
-	dprintf("%"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32"\n", my_idx, cur_idx, cur_nr, copy_idx);
+	if (copy_idx < 0) {
+		eprintf("shouldn't happen\n");
+		goto out;
+	}
 
-	ret = __recover_one(rw, old_entry, old_nr, cur_entry, cur_nr,
+	dprintf("%"PRIu32", %"PRIu32", %"PRIu32"\n", cur_idx, cur_nr_nodes, copy_idx);
+
+	ret = __recover_one(rw, old_vnodes, old_nr_vnodes, cur_vnodes, cur_nr_vnodes,
 			    cur_idx, copy_idx, epoch, epoch - 1, oid,
 			    buf, SD_DATA_OBJ_SIZE);
 	if (ret == 0)
@@ -1358,8 +1405,8 @@ static void recover_one(struct work *work, int idx)
 	for (i = 0; i < sys->nr_sobjs; i++) {
 		if (i == copy_idx)
 			continue;
-		ret = __recover_one(rw, old_entry, old_nr,
-				    cur_entry, cur_nr, cur_idx, i,
+		ret = __recover_one(rw, old_vnodes, old_nr_vnodes,
+				    cur_vnodes, cur_nr_vnodes, cur_idx, i,
 				    epoch, epoch - 1, oid, buf, SD_DATA_OBJ_SIZE);
 		if (ret == 0)
 			goto out;
@@ -1369,27 +1416,22 @@ fail:
 out:
 	if (buf)
 		free(buf);
-
-	if (!rw->retry)
-		rw->done++;
 }
 
 static struct recovery_work *suspended_recovery_work;
-static uint64_t recovering_oid;
 
 static void __start_recovery(struct work *work, int idx);
 
 static void recover_timer(void *data)
 {
 	struct recovery_work *rw = (struct recovery_work *)data;
-	uint64_t oid = *(((uint64_t *)rw->buf) + rw->done);
+	uint64_t oid = rw->oids[rw->done];
 
 	if (is_access_to_busy_objects(oid)) {
 		suspended_recovery_work = rw;
 		return;
 	}
 
-	recovering_oid = oid;
 	queue_work(&rw->work);
 }
 
@@ -1403,21 +1445,20 @@ void resume_recovery_work(void)
 
 	rw = suspended_recovery_work;
 
-	oid =  *(((uint64_t *)rw->buf) + rw->done);
+	oid =  rw->oids[rw->done];
 	if (is_access_to_busy_objects(oid))
 		return;
 
 	suspended_recovery_work = NULL;
-	recovering_oid = oid;
 	queue_work(&rw->work);
 }
 
 int is_recoverying_oid(uint64_t oid)
 {
 	uint64_t hval = fnv_64a_buf(&oid, sizeof(uint64_t), FNV1A_64_INIT);
-	uint64_t recovering_hval = fnv_64a_buf(&recovering_oid, sizeof(uint64_t), FNV1A_64_INIT);
+	uint64_t min_hval;
 	struct recovery_work *rw = recovering_work;
-	int ret, fd;
+	int ret, fd, i;
 
 	if (oid == 0)
 		return 0;
@@ -1425,9 +1466,14 @@ int is_recoverying_oid(uint64_t oid)
 	if (!rw)
 		return 0; /* there is no thread working for object recovery */
 
+	min_hval = fnv_64a_buf(&rw->oids[rw->done + rw->nr_blocking], sizeof(uint64_t), FNV1A_64_INIT);
+
 	if (before(rw->epoch, sys->epoch))
 		return 1;
 
+	if (rw->state == RW_INIT)
+		return 1;
+
 	fd = ob_open(sys->epoch, oid, 0, &ret);
 	if (fd != -1) {
 		dprintf("the object %" PRIx64 " is already recoverd\n", oid);
@@ -1435,11 +1481,27 @@ int is_recoverying_oid(uint64_t oid)
 		return 0;
 	}
 
-	if (recovering_hval <= hval) {
-		if (bsearch(&oid, ((uint64_t *)rw->buf) + rw->done,
-			    rw->count - rw->done, sizeof(oid), obj_cmp)) {
+	/* the first 'rw->nr_blocking' objects were already scheduled to be done earlier */
+	for (i = 0; i < rw->nr_blocking; i++)
+		if (rw->oids[rw->done + i] == oid)
+			return 1;
+
+	if (min_hval <= hval) {
+		uint64_t *p;
+		p = bsearch(&oid, rw->oids + rw->done + rw->nr_blocking,
+			    rw->count - rw->done - rw->nr_blocking, sizeof(oid), obj_cmp);
+		if (p) {
 			dprintf("recover the object %" PRIx64 " first\n", oid);
-			blocking_oid = oid;
+			if (rw->nr_blocking == 0)
+				rw->nr_blocking = 1; /* the first oid may be processed now */
+			if (p > rw->oids + rw->done + rw->nr_blocking) {
+				/* this object should be recovered earlier */
+				memmove(rw->oids + rw->done + rw->nr_blocking + 1,
+					rw->oids + rw->done + rw->nr_blocking,
+					sizeof(uint64_t) * (p - (rw->oids + rw->done + rw->nr_blocking)));
+				rw->oids[rw->done + rw->nr_blocking] = oid;
+				rw->nr_blocking++;
+			}
 			return 1;
 		}
 	}
@@ -1451,7 +1513,17 @@ int is_recoverying_oid(uint64_t oid)
 static void recover_done(struct work *work, int idx)
 {
 	struct recovery_work *rw = container_of(work, struct recovery_work, work);
-	uint64_t oid = *(((uint64_t *)rw->buf) + rw->done);
+	uint64_t oid;
+
+	if (rw->state == RW_INIT)
+		rw->state = RW_RUN;
+	else if (!rw->retry) {
+		rw->done++;
+		if (rw->nr_blocking > 0)
+			rw->nr_blocking--;
+	}
+
+	oid = rw->oids[rw->done];
 
 	if (rw->retry && list_empty(&recovery_work_list)) {
 		rw->retry = 0;
@@ -1462,8 +1534,6 @@ static void recover_done(struct work *work, int idx)
 		return;
 	}
 
-	blocking_oid = 0;
-
 	if (rw->done < rw->count && list_empty(&recovery_work_list)) {
 		rw->work.fn = recover_one;
 
@@ -1471,20 +1541,18 @@ static void recover_done(struct work *work, int idx)
 			suspended_recovery_work = rw;
 			return;
 		}
-		recovering_oid = oid;
 		resume_pending_requests();
 		queue_work(&rw->work);
 		return;
 	}
 
 	dprintf("recovery done, %"PRIu32"\n", rw->epoch);
-	recovering_oid = 0;
 	recovering_work = NULL;
 
 	sys->recovered_epoch = rw->epoch;
 	resume_pending_requests();
 
-	free(rw->buf);
+	free(rw->oids);
 	free(rw);
 
 	if (!list_empty(&recovery_work_list)) {
@@ -1498,12 +1566,10 @@ static void recover_done(struct work *work, int idx)
 	}
 }
 
-static int __fill_obj_list(struct recovery_work *rw,
-			   struct sheepdog_node_list_entry *e,
-			   uint64_t start_hash, uint64_t end_hash, uint64_t *done_hash)
+static int __fill_obj_list(struct sheepdog_node_list_entry *e, uint32_t epoch,
+			   uint8_t *buf, size_t buf_size)
 {
 	int fd, ret;
-	uint32_t epoch = rw->epoch;
 	unsigned wlen, rlen;
 	char name[128];
 	struct sd_list_req hdr;
@@ -1520,72 +1586,111 @@ static int __fill_obj_list(struct recovery_work *rw,
 	}
 
 	wlen = 0;
-	rlen = (1 << 20) - (rw->count * sizeof(uint64_t));
+	rlen = buf_size;
 
 	memset(&hdr, 0, sizeof(hdr));
 	hdr.opcode = SD_OP_GET_OBJ_LIST;
 	/* we don't need to set epoch */
 	hdr.epoch = epoch;
-	hdr.start = start_hash;
-	hdr.end = end_hash;
 	hdr.tgt_epoch = epoch - 1;
 	hdr.flags = 0;
 	hdr.data_length = rlen;
 
-	dprintf("%016"PRIx64", %016"PRIx64"\n", hdr.start, hdr.end);
-
-	ret = exec_req(fd, (struct sd_req *)&hdr, rw->buf + rw->count * sizeof(uint64_t), &wlen, &rlen);
+	ret = exec_req(fd, (struct sd_req *)&hdr, buf, &wlen, &rlen);
 
 	close(fd);
 
 	rsp = (struct sd_list_rsp *)&hdr;
 
 	if (ret || rsp->result != SD_RES_SUCCESS) {
-		rw->retry = 1;
-		*done_hash = end_hash;
 		eprintf("try again, %"PRIu32", %"PRIu32"\n", ret, rsp->result);
-		return 0;
+		return -1;
 	}
 
 	dprintf("%"PRIu32"\n", rsp->data_length);
 
-	if (rsp->data_length)
-		rw->count += rsp->data_length / sizeof(uint64_t);
-
-	*done_hash = rsp->next;
-
-	return 0;
+	return rsp->data_length / sizeof(uint64_t);
 }
 
+static int merge_objlist(struct sheepdog_vnode_list_entry *entries, int nr_entries,
+			 uint64_t *list1, int nr_list1,
+			 uint64_t *list2, int nr_list2, int nr_objs)
+{
+	int i, j, idx;
+
+	for (i = 0; i < nr_list2; i++) {
+		for (j = 0; j < nr_objs; j++) {
+			idx = obj_to_sheep(entries, nr_entries, list2[i], j);
+			if (is_myself(entries[idx].addr, entries[idx].port))
+				break;
+		}
+		if (j == nr_objs)
+			continue;
+
+		if (bsearch(list2 + i, list1, nr_list1, sizeof(*list1), obj_cmp))
+			continue;
+
+		list1[nr_list1++] = list2[i];
+	}
+
+	qsort(list1, nr_list1, sizeof(*list1), obj_cmp);
+
+	return nr_list1;
+}
+
+#define MAX_RETRY_CNT  6
+
 static int fill_obj_list(struct recovery_work *rw,
 			 struct sheepdog_node_list_entry *old_entry, int old_nr,
 			 struct sheepdog_node_list_entry *cur_entry, int cur_nr,
-			 uint64_t start_hval, uint64_t end_hval, int nr_objs)
+			 int nr_objs)
 {
-	int i, idx, old_idx, cur_idx;
-	uint64_t hval, done_hval = end_hval;
+	int i, j;
+	uint8_t *buf = NULL;
+	size_t buf_size = SD_DATA_OBJ_SIZE; /* FIXME */
+	struct sheepdog_vnode_list_entry vnodes[SD_MAX_VNODES];
+	int nr_vnodes, retry_cnt = 0;
 
-	hval = start_hval;
-again:
-	old_idx = hval_to_sheep(old_entry, old_nr, hval + 1, 0);
-	cur_idx = hval_to_sheep(cur_entry, cur_nr, hval + 1, 0);
+	buf = malloc(buf_size);
+	if (!buf)
+		goto fail;
 
-	for (i = 0; i < nr_objs; i++) {
-		idx = find_tgt_node(old_entry, old_nr, old_idx, cur_entry, cur_nr, cur_idx, i);
-		dprintf("%"PRIu32", %"PRIu32"\n", idx, i);
-		if (__fill_obj_list(rw, old_entry + idx, hval, end_hval, &done_hval) == 0)
-			break;
-	}
-	if (i == nr_objs)
-		return -1;
+	nr_vnodes = nodes_to_vnodes(cur_entry, cur_nr, vnodes);
+	for (i = 0; i < cur_nr; i++) {
+		int nr;
 
-	if (done_hval != end_hval) {
-		dprintf("%"PRIx64", %"PRIx64"\n", done_hval, end_hval);
-		hval = done_hval;
-		goto again;
+		for (j = 0; j < old_nr; j++)
+			if (node_cmp(cur_entry + i, old_entry + j) == 0)
+				break;
+
+		if (j == old_nr)
+			/* cur_entry[i] doesn't have a list file */
+			continue;
+
+	retry:
+		nr  = __fill_obj_list(cur_entry + i, rw->epoch, buf, buf_size);
+		if (nr < 0) {
+			retry_cnt++;
+			if (retry_cnt > MAX_RETRY_CNT) {
+				eprintf("failed to get object list\n");
+				eprintf("some objects may be lost\n");
+				continue;
+			} else {
+				dprintf("retry get object list\n");
+				sleep(1);
+				goto retry;
+			}
+		}
+		rw->count = merge_objlist(vnodes, nr_vnodes, rw->oids,
+					  rw->count, (uint64_t *)buf, nr, nr_objs);
 	}
 
+	free(buf);
 	return 0;
+fail:
+	free(buf);
+	rw->retry = 1;
+	return -1;
 }
 
 static void __start_recovery(struct work *work, int idx)
@@ -1595,9 +1700,7 @@ static void __start_recovery(struct work *work, int idx)
 	struct sheepdog_node_list_entry old_entry[SD_MAX_NODES],
 		cur_entry[SD_MAX_NODES];
 	int old_nr, cur_nr, nr_objs;
-	int my_idx = -1;
-	int i, fd;
-	uint64_t start_hash, end_hash;
+	int fd;
 	char path[PATH_MAX], tmp_path[PATH_MAX];
 	int ret;
 
@@ -1623,18 +1726,7 @@ static void __start_recovery(struct work *work, int idx)
 	if (!nr_objs)
 		goto fail;
 
-	for (i = 0; i < cur_nr; i++) {
-		if (cur_entry[i].id == sys->this_node.id) {
-			my_idx = i;
-			break;
-		}
-	}
-	start_hash = cur_entry[(my_idx - nr_objs + cur_nr) % cur_nr].id;
-	end_hash = cur_entry[my_idx].id;
-
-	dprintf("fill obj list (from 0x%"PRIx64" to 0x%"PRIx64")\n", start_hash, end_hash);
-	if (fill_obj_list(rw, old_entry, old_nr, cur_entry, cur_nr,
-			  start_hash, end_hash, nr_objs) != 0) {
+	if (fill_obj_list(rw, old_entry, old_nr, cur_entry, cur_nr, nr_objs) != 0) {
 		eprintf("fatal recovery error\n");
 		goto fail;
 	}
@@ -1642,8 +1734,6 @@ static void __start_recovery(struct work *work, int idx)
 	if (rw->retry)
 		goto fail;
 
-	qsort(rw->buf, rw->count, sizeof(uint64_t), obj_cmp);
-
 	snprintf(path, sizeof(path), "%s%08u/list", obj_path, epoch);
 	snprintf(tmp_path, sizeof(tmp_path), "%s%08u/list.tmp", obj_path, epoch);
 
@@ -1653,7 +1743,7 @@ static void __start_recovery(struct work *work, int idx)
 		eprintf("failed to open %s, %s, %m\n", tmp_path, strerror(errno));
 		goto fail;
 	}
-	ret = write(fd, rw->buf, sizeof(uint64_t) * rw->count);
+	ret = write(fd, rw->oids, sizeof(*rw->oids) * rw->count);
 	if (ret != sizeof(uint64_t) * rw->count) {
 		eprintf("failed to write to %s, %m\n", tmp_path);
 		close(fd);
@@ -1682,7 +1772,8 @@ int start_recovery(uint32_t epoch)
 	if (!rw)
 		return -1;
 
-	rw->buf = malloc(1 << 20); /* FIXME */
+	rw->state = RW_INIT;
+	rw->oids = malloc(1 << 20); /* FIXME */
 	rw->epoch = epoch;
 	rw->count = 0;
 
@@ -1761,16 +1852,6 @@ static int attr(char *path, const char *name, void *var, int len, int set)
 	return SD_RES_SUCCESS;
 }
 
-int set_nodeid(uint64_t nodeid)
-{
-	return attr(epoch_path, ANAME_NODEID, &nodeid, sizeof(nodeid), 1);
-}
-
-int get_nodeid(uint64_t *nodeid)
-{
-	return attr(epoch_path, ANAME_NODEID, nodeid, sizeof(*nodeid), 0);
-}
-
 static int init_base_path(const char *d)
 {
 	int new = 0;
@@ -1831,7 +1912,7 @@ static int init_epoch_path(const char *base_path)
 
 		oid = strtoull(dent->d_name, NULL, 16);
 
-		if (is_data_obj(oid))
+		if (!is_vdi_obj(oid))
 			continue;
 
 		vprintf(SDOG_DEBUG "found the vdi obj, %" PRIx64 "\n", oid);
